# Project: finmarketpy project
# Filename: learn_network_structure
# Objective: compute a network graph for a group of asset return time series
# Created: 2019-11-02 12:05
# Version: 0.0
# Author: FS

__author__ = 'fs'

#
# Copyright 2016-2020 Cuemacro - https://www.cuemacro.com / @cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import numpy as np
from sklearn import cluster, covariance, manifold


def learn_network_structure(ts_returns_data, names, alphas=4, cv=5, mode='cd',
							assume_centered = False,
							n_components=2, n_neighbors=5,
							eigen_solver="dense", method='standard',
							neighbors_algorithm="auto",
							random_state = None, n_jobs=None,
							standardise=False):
	"""

	Parameters
	----------
	ts_returns_data : array-like of shape [n_samples, n_instruments]
	                  time series matrix of returns

	names : array-like of shape [n_samples, 1]
	        Individual names of the financial instrument

	alphas : int or positive float, optional
	         Number of points on the grids to be used

	cv : int, optional
	     Number of folds for cross-validation splitting strategy

	mode : str, optional
	       Solver to use to compute the graph

	assume_centered : bool, optional
                      Centre the data if False.

	n_components : int
	               Number of components for the manifold

	n_neighbors: int
                 Number of neighbours to consider for each point

	eigen_solver : str
	               Algorithm to compute eigenvalues

	method : str
             Algorithm to use for local linear embedding
	neighbors_algorithm : str
	                      Algorithm to use for nearest neighbours search

	random_state : int, RandomState instance or None, optional
	               If int, random_state is the seed used by the random number generator.
	               If RandomState instance, random_state is the random number generator.
	               If None, the random number generator is the RandomState instance used by np.random.
	               Used when eigen_solver == ‘arpack’

	n_jobs : int or None, optional
	         number of parallel jobs to run

	standardise : bool
	              standardise data if True

	Returns : sklearn.covariance.graph_lasso_.GraphicalLassoCV

              sklearn.manifold.locally_linear.LocallyLinearEmbedding

              array-like of shape [n_components, n_instruments]
              Transformed embedding vectors

              array-like of shape [n_instruments, 1]
              numeric identifier of each cluster



	-------
	"""

	if not isinstance(ts_returns_data, (np.ndarray, np.generic)):
		raise TypeError("ts_returns_data must be of class ndarray")

	# learn graphical structure
	edge_model = covariance.GraphicalLassoCV(alphas=alphas, cv=cv, mode=mode,
											 assume_centered=assume_centered)
	edge_model.fit(ts_returns_data)

	# cluster using affinity propagation
	_, labels = cluster.affinity_propagation(edge_model.covariance_)
	n_labels = labels.max()
	for i in range(n_labels + 1):
		print('Cluster %i: %s' % ((i + 1), ', '.join(names[labels == i])))

	# find low-dimension embedding - useful for 2D plane visualisation
	node_position_model = manifold.LocallyLinearEmbedding(
			n_components=n_components, eigen_solver=eigen_solver,
			n_neighbors=n_neighbors, method=method,
			neighbors_algorithm=neighbors_algorithm,
			random_state=random_state, n_jobs=n_jobs)
	embedding = node_position_model.fit_transform(ts_returns_data.T).T

	if standardise:
		# standardise returns
		standard_ret = ts_returns_data.copy()
		standard_ret /= ts_returns_data.std(axis=0)

		# learn graph model
		edge_model.fit(standard_ret)

		# cluster using affinity propagation
		_, labels = cluster.affinity_propagation(edge_model.covariance_)
		n_labels = labels.max()
		for i in range(n_labels + 1):
			print('Cluster %i: %s' % ((i + 1), ', '.join(names[labels == i])))

		# find low-dimension embedding - useful for 2D plane visualisation
		node_position_model = manifold.LocallyLinearEmbedding(
				n_components=n_components, eigen_solver=eigen_solver,
				n_neighbors=n_neighbors, method=method,
				neighbors_algorithm=neighbors_algorithm,
				random_state=random_state, n_jobs=n_jobs)
		embedding = node_position_model.fit_transform(ts_returns_data.T).T

	return edge_model, node_position_model, embedding, labels



__author__ = "saeedamen"

#
# Copyright 2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Shows how to use finmarketpy to price FX options (uses FinancePy underneath - 
it is recommended you pull the latest version of FinancePy from GitHub).

Note, you will need to have a Bloomberg terminal (with blpapi Python library) 
to download the FX market data in order to plot these vol surface (FX spot, 
FX forwards, FX implied_vol volatility quotes and deposits)
"""

import pandas as pd

# For plotting
from chartpy import Chart, Style

# For loading market data
from findatapy.market import Market, MarketDataGenerator, MarketDataRequest

from findatapy.util.loggermanager import LoggerManager

from finmarketpy.curve.rates.fxforwardspricer import FXForwardsPricer
from finmarketpy.curve.volatility.fxvolsurface import FXVolSurface
from finmarketpy.curve.volatility.fxoptionspricer import FXOptionsPricer

logger = LoggerManager().getLogger(__name__)

chart = Chart(engine="plotly")

market = Market(market_data_generator=MarketDataGenerator())

# Choose run_example = 0 for everything
# run_example = 1 - price GBPUSD options
# run_example = 2 - price USDJPY options
# run_example = 3 - price AUDUSD options
# run_example = 4 - more pricing of AUDUSD options
# run_example = 5 - pricing of EURUSD options
# run_example = 6 - another USDJPY option
# run_example = 7 - price USDBRL options

run_example = 0

if __name__ == "__main__":

    # Fetch market data for pricing GBPUSD FX options over Brexit vote
    # (ie. FX spot, FX forwards, FX deposits and FX vol quotes)
    # Construct volatility surface using FinancePy library underneath,
    # using polynomial interpolation and then price some options over these
    # dates eg. atm, 25d-call etc.
    if run_example == 1 or run_example == 0:
        horizon_date = "23 Jun 2016"
        cross = "GBPUSD"

        # Download the whole all market data for GBPUSD for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=horizon_date,
                                       finish_date=horizon_date,
                                       data_source="bloomberg", cut="NYC",
                                       category="fx-vol-market",
                                       tickers=cross,
                                       base_depos_currencies=[cross[0:3],
                                                              cross[3:6]],
                                       cache_algo="cache_algo_return")

        df = market.fetch_market(md_request)

        fx_vol_surface = FXVolSurface(market_df=df, asset=cross)

        fx_op = FXOptionsPricer(fx_vol_surface=fx_vol_surface)

        # Price several different options

        print("atm 1M european call")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), "atm",
                                     contract_type="european-call",
                                     tenor="1M").to_string())

        print("25d 1W european put")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date),
                                     "25d-otm", contract_type="european-put",
                                     tenor="1W").to_string())

        # Try a broken date 12D option (note, for broken dates, currently doesn"t interpolate key strikes)
        # Specify expiry date instead of the tenor for broken dates
        print("1.50 12D european call")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), 1.50,
                                     expiry_date=pd.Timestamp(
                                         horizon_date) + pd.Timedelta(days=12),
                                     contract_type="european-call").to_string())

    ###### Fetch market data for pricing USDJPY FX options over Brexit vote (ie. FX spot, FX forwards, FX deposits and FX vol quotes)
    ###### Construct volatility surface using FinancePy library underneath, using polynomial interpolation
    ###### Then price a series of 1W ATM call options
    if run_example == 2 or run_example == 0:
        start_date = "02 Nov 2020";
        finish_date = "05 Nov 2020"
        horizon_date = pd.bdate_range(start_date, finish_date, freq="B")

        cross = "USDJPY"

        # Download the whole all market data for GBPUSD for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=start_date,
                                       finish_date=finish_date,
                                       data_source="bloomberg", cut="NYC",
                                       category="fx-vol-market",
                                       tickers=cross,
                                       cache_algo="cache_algo_return",
                                       base_depos_currencies=[cross[0:3],
                                                              cross[3:6]])

        df = market.fetch_market(md_request)

        # Skip 3W/4M because this particular close (NYC) doesn"t have that in USDJPY market data
        tenors = ["ON", "1W", "2W", "1M", "2M", "3M", "6M", "9M", "1Y", "2Y",
                  "3Y"]
        fx_vol_surface = FXVolSurface(market_df=df, asset=cross, tenors=tenors)

        fx_op = FXOptionsPricer(fx_vol_surface=fx_vol_surface)

        print("atm 1W european put")
        print(fx_op.price_instrument(cross, horizon_date, "atm",
                                     contract_type="european-put",
                                     tenor="1W", depo_tenor="1W").to_string())

        print("25d 3M european call")
        print(fx_op.price_instrument(cross, horizon_date, "25d-otm",
                                     contract_type="european-call",
                                     tenor="3M", depo_tenor="3M").to_string())

        print("10d 1M european put")
        print(fx_op.price_instrument(cross, horizon_date, "10d-otm",
                                     contract_type="european-put",
                                     tenor="1M", depo_tenor="1M").to_string())

    ###### Fetch market data for pricing AUDUSD options on 18 Apr 2007, just before credit crisis
    ###### Construct volatility surface using FinancePy library underneath, using polynomial interpolation and
    ###### Then price some options over these dates eg. atm, 25d-call etc.
    if run_example == 3 or run_example == 0:
        horizon_date = "18 Apr 2007"
        cross = "AUDUSD"

        # Download the whole all market data for GBPUSD for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=horizon_date,
                                       finish_date=horizon_date,
                                       data_source="bloomberg", cut="LDN",
                                       category="fx-vol-market",
                                       tickers=cross,
                                       base_depos_currencies=[cross[0:3],
                                                              cross[3:6]],
                                       cache_algo="cache_algo_return")

        df = market.fetch_market(md_request)

        fx_vol_surface = FXVolSurface(market_df=df, asset=cross,
                                      tenors=["ON", "1W", "1M"])

        fx_op = FXOptionsPricer(fx_vol_surface=fx_vol_surface)

        # Try a broken date 15D option (note, for broken dates, currently doesn"t interpolate key strikes)
        # Specify expiry date instead of the tenor for broken dates
        print("atm 15D european call")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), 0.8124,
                                     expiry_date=pd.Timestamp(
                                         horizon_date) + pd.Timedelta(days=15),
                                     contract_type="european-call").to_string())

    ###### Fetch market data for pricing AUDUSD options during start of 2008 Credit Crisis
    if run_example == 4 or run_example == 0:
        horizon_date = "17 Aug 2007"
        cross = "AUDUSD"

        # Download the whole all market data for GBPUSD for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=horizon_date,
                                       finish_date=horizon_date,
                                       data_source="bloomberg", cut="BGN",
                                       category="fx-vol-market",
                                       tickers=cross,
                                       base_depos_currencies=[cross[0:3],
                                                              cross[3:6]],
                                       cache_algo="cache_algo_return")

        df = market.fetch_market(md_request)

        fx_vol_surface = FXVolSurface(market_df=df, asset=cross,
                                      tenors=["1W", "1M", "3M"])
        fx_vol_surface.build_vol_surface(pd.Timestamp(horizon_date))

        fx_op = FXOptionsPricer(fx_vol_surface=fx_vol_surface)

        # Price several different options

        # Try a broken date 15D option (note, for broken dates, currently doesn"t interpolate key strikes)
        # Specify expiry date instead of the tenor for broken dates
        print("atm 15D european call")

        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), 0.8535,
                                     expiry_date=pd.Timestamp("05 Sep 2007"),
                                     contract_type="european-call").to_string())

    ###### Fetch market data for pricing EURUSD options during start of 2006
    if run_example == 5 or run_example == 0:
        horizon_date = "04 Jan 2006"
        cross = "EURUSD"

        # Download the whole all market data for GBPUSD for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=horizon_date,
                                       finish_date=horizon_date,
                                       data_source="bloomberg", cut="BGN",
                                       category="fx-vol-market",
                                       tickers=cross,
                                       base_depos_currencies=[cross[0:3],
                                                              cross[3:6]],
                                       cache_algo="cache_algo_return")

        df = market.fetch_market(md_request)

        fx_vol_surface = FXVolSurface(market_df=df, asset=cross,
                                      tenors=["1W", "1M", "3M"])

        fx_op = FXOptionsPricer(fx_vol_surface=fx_vol_surface)

        # Price several different options

        # Try a broken date 15D option (note, for broken dates, currently doesn"t interpolate key strikes)
        # Specify expiry date instead of the tenor for broken dates
        print("atm 1W european call")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), "atm",
                                     tenor="1W", depo_tenor="1W",
                                     contract_type="european-call").to_string())

    ###### Fetch market data for pricing USDJPY ATM 1W
    if run_example == 6 or run_example == 0:
        horizon_date = "30 March 2007"
        cross = "USDJPY"

        # Download the whole all market data for GBPUSD for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=horizon_date,
                                       finish_date=horizon_date,
                                       data_source="bloomberg", cut="LDN",
                                       category="fx-vol-market",
                                       fx_vol_tenor=["1W"],
                                       tickers=cross,
                                       base_depos_currencies=[cross[0:3],
                                                              cross[3:6]],
                                       cache_algo="cache_algo_return")

        df = market.fetch_market(md_request)

        fx_vol_surface = FXVolSurface(market_df=df, asset=cross, tenors=["1W"],
                                      solver="nelmer-mead-numba")

        fx_op = FXOptionsPricer(fx_vol_surface=fx_vol_surface)

        market_df = fx_vol_surface.get_all_market_data()

        # Print 1W data
        print(market_df[[x for x in market_df.columns if "1W" in x]][
                  market_df.index == horizon_date].to_string())

        # Print ATM vol
        fx_vol_surface.build_vol_surface(horizon_date)
        fx_vol_surface.extract_vol_surface(num_strike_intervals=None)

        print("ATM vol " + str(fx_vol_surface.get_atm_vol(tenor="1W")))

        # Specify expiry date instead of the tenor for broken dates
        print("atm 1W european straddle")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), "atm",
                                     tenor="1W", depo_tenor="1W",
                                     contract_type="european-straddle").to_string())

    ###### Price USDBRL option around 2018 2nd round of presidential election
    if run_example == 7 or run_example == 0:
        horizon_date = "26 Oct 2018"
        cross = "USDBRL"
        non_usd = "BRL"

        # Download the whole all market data for USDBRL for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=horizon_date,
                                       finish_date=horizon_date,
                                       data_source="bloomberg", cut="NYC",
                                       category="fx-vol-market",
                                       tickers=cross,
                                       base_depos_currencies=[cross[0:3]],
                                       cache_algo="cache_algo_return")

        df = market.fetch_market(md_request)

        # Compute implied deposit BRL 1M from USDBRL forwards (and USD 1M depo)
        fx_forwards_price = FXForwardsPricer()

        implied_depo_df = fx_forwards_price.calculate_implied_depo(cross,
                                                                   non_usd,
                                                                   market_df=df,
                                                                   fx_forwards_tenor=[
                                                                       "1W",
                                                                       "1M"],
                                                                   depo_tenor=[
                                                                       "1W",
                                                                       "1M"])

        implied_depo_df.columns = [x.replace("-implied-depo", "") for x in
                                   implied_depo_df.columns]
        df = df.join(implied_depo_df, how="left")

        # USDBRL quoted ATMF implied vol (as opposed to delta neutral) hence "fwd" parameter
        fx_op = FXOptionsPricer(
            fx_vol_surface=FXVolSurface(market_df=df, asset=cross,
                                        atm_method="fwd", depo_tenor="1M"))

        # Price several different options
        print(df)

        print("atm 1M european put")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), "atm",
                                     contract_type="european-put",
                                     tenor="1M").to_string())

        # TODO: calendar around election results in slightly different pricing
        # print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), "25d-otm", contract_type="european-put", tenor="1W").to_string())
        # print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), 3.5724, contract_type="european-put", expiry_date=pd.Timestamp("2 Nov 2018")).to_string())

    ###### Price GBPUSD option around Brexit with unquoted deltas
    if run_example == 8 or run_example == 0:
        horizon_date = "23 Jun 2016"
        cross = "GBPUSD"

        # Download the whole all market data for GBPUSD for pricing options (vol surface)
        md_request = MarketDataRequest(start_date=horizon_date,
                                       finish_date=horizon_date,
                                       data_source="bloomberg", cut="NYC",
                                       category="fx-vol-market",
                                       tickers=cross,
                                       base_depos_currencies=[cross[0:3],
                                                              cross[3:6]],
                                       cache_algo="cache_algo_return")

        df = market.fetch_market(md_request)

        fx_vol_surface = FXVolSurface(market_df=df, asset=cross)

        fx_op = FXOptionsPricer(fx_vol_surface=fx_vol_surface)

        # Price several different options
        print("atm 1M european call")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date), "atm",
                                     contract_type="european-call",
                                     tenor="1M").to_string())

        print("25d 1W european put")
        print(fx_op.price_instrument(cross, pd.Timestamp(horizon_date),
                                     "25d-otm", contract_type="european-put",
                                     tenor="1W").to_string())



__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro - https://www.cuemacro.com / @cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

from chartpy import Chart, Style, ChartConstants

from findatapy.util.dataconstants import DataConstants
from findatapy.market import Market, MarketDataRequest, MarketDataGenerator
from findatapy.timeseries import Calculations, RetStats

import datetime

from datetime import timedelta

dataconstants = DataConstants()

class QuickChart(object):
    """Displays charts from downloaded data, in a single line code call. Ideal for quickly generating charts from sources
    including Bloomberg, Quandl, ALFRED/FRED etc.

    """

    def __init__(self, engine='plotly', data_source='bloomberg', market_data_generator=MarketDataGenerator()):
        self._chart = Chart(engine=engine)
        self._market = Market(market_data_generator=market_data_generator)
        self._data_source = data_source

    def plot_chart(self, tickers=None, tickers_rhs=None, start_date=None, finish_date=None,
                   chart_file=None, chart_type='line', title='',
                   fields={'close' : 'PX_LAST'}, freq='daily', source='Web', brand_label='Cuemacro', display_brand_label=True,
                   reindex=False, additive_index=False, yoy=False, plotly_plot_mode='offline_png',
                   height=400, width=600, scale_factor=-1,
                   quandl_api_key=dataconstants.quandl_api_key,
                   fred_api_key=dataconstants.fred_api_key,
                   alpha_vantage_api_key=dataconstants.alpha_vantage_api_key, df=None):

        if start_date is None:
            start_date = datetime.datetime.utcnow().date() - timedelta(days=60)
        if finish_date is None:
            finish_date = datetime.datetime.utcnow()

        if isinstance(tickers, str):
            tickers = {tickers : tickers}
        elif isinstance(tickers, list):
            tickers_dict = {}

            for t in tickers:
                tickers_dict[t] = t

            tickers = tickers_dict

        if tickers_rhs is not None:
            if isinstance(tickers_rhs, str):
                tickers_rhs = {tickers_rhs: tickers_rhs}
            elif isinstance(tickers, list):
                tickers_rhs_dict = {}

                for t in tickers_rhs:
                    tickers_rhs_dict[t] = t

                tickers_rhs = tickers_rhs_dict

            tickers.update(tickers_rhs)
        else:
            tickers_rhs = {}

        if df is None:
            md_request = MarketDataRequest(start_date=start_date, finish_date=finish_date,
                                           freq=freq,
                                           data_source=self._data_source,
                                           tickers=list(tickers.keys()), vendor_tickers=list(tickers.values()),
                                           fields=list(fields.keys()),
                                           vendor_fields=list(fields.values()),
                                           quandl_api_key=quandl_api_key,
                                           fred_api_key=fred_api_key,
                                           alpha_vantage_api_key=alpha_vantage_api_key)

            df = self._market.fetch_market(md_request=md_request)

        df = df.fillna(method='ffill')
        df.columns = [x.split('.')[0] for x in df.columns]

        style = Style(title=title, chart_type=chart_type, html_file_output=chart_file, scale_factor=scale_factor,
                      height=height, width=width, file_output=datetime.date.today().strftime("%Y%m%d") + " " + title + ".png",
                      plotly_plot_mode=plotly_plot_mode, source=source, brand_label=brand_label,
                      display_brand_label=display_brand_label)

        if reindex:
            df = Calculations().create_mult_index_from_prices(df)

            style.y_title = 'Reindexed from 100'

        if additive_index:
            df = (df - df.shift(1)).cumsum()

            style.y_title = 'Additive changes from 0'

        if yoy:
            if freq == 'daily':
                obs_in_year = 252
            elif freq == 'intraday':
                obs_in_year = 1440

            df_rets = Calculations().calculate_returns(df)
            df = Calculations().average_by_annualised_year(df_rets, obs_in_year=obs_in_year) * 100

            style.y_title = 'Annualized % YoY'

        if list(tickers_rhs.keys()) != []:
            style.y_axis_2_series=list(tickers_rhs.keys())
            style.y_axis_2_showgrid = False
            style.y_axis_showgrid = False

        return self._chart.plot(df, style=style), df

    def plot_chart_with_ret_stats(self, df=None, chart_file=None, chart_type='line', title='',
                   height=400, width=600, scale_factor=-1,
                   ann_factor=252, source='Web', brand_label='Cuemacro', display_brand_label=True,
                   plotly_plot_mode='offline_png'):

        style = Style(title=title, chart_type=chart_type, html_file_output=chart_file, scale_factor=scale_factor,
                      height=height, width=width, file_output=datetime.date.today().strftime("%Y%m%d") + " " + title + ".png",
                      plotly_plot_mode=plotly_plot_mode, source=source, brand_label=brand_label,
                      display_brand_label=display_brand_label)

        returns_df = df / df.shift(1) - 1.0

        ret_stats = RetStats(returns_df=returns_df, ann_factor=ann_factor)

        df.columns = ret_stats.summary()

        return self._chart.plot(df, style=style), df





__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro - https://www.cuemacro.com / @cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import abc

import numpy as np

ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

from financepy.utils.date import Date

class AbstractVolSurface(ABC):
    """Holds data for an asset class vol surface

    """

    @abc.abstractmethod
    def build_vol_surface(self):
        return

    @abc.abstractmethod
    def extract_vol_surface(self):
        return

    def _extremes(self, min, max, data):
        if min is None:
            min = np.min(data)
        else:
            new_min = np.min(data)

            if new_min < min:
                min = new_min

        if max is None:
            max = np.max(data)
        else:
            new_max = np.max(data)

            if new_max > max:
                max = new_max

        return min, max

    def extract_vol_surface_across_dates(self, dates,
                                         num_strike_intervals=60, vol_surface_type='vol_surface_strike_space',
                                         reverse_plot=True):

        vol_surface_dict = {}

        min_x = None
        max_x = None
        min_z = None
        max_z = None

        for i in range(0, len(dates)):

            self.build_vol_surface(dates[i])

            # Note for unstable vol surface dates (eg. over Brexit date) you may need to increase tolerance
            # in FinancePy
            # in FinFXVolSurface.buildVolSurface method
            df_vol_surface = self.extract_vol_surface(num_strike_intervals=num_strike_intervals)[vol_surface_type]

            if reverse_plot:
                # Reverse order of tenors to match the way BBG plots it
                vol_surface_dict[dates[i]] = df_vol_surface.iloc[:, ::-1]
            else:
                vol_surface_dict[dates[i]] = df_vol_surface

            # x_axis = strike - index
            # y_axis = tenor - columns
            # z_axis = implied_vol vol - values
            min_x, max_x = self._extremes(min_x, max_x, df_vol_surface.index.values)
            min_z, max_z = self._extremes(min_z, max_z, df_vol_surface.values)

        extremes_dict = {'min_x': min_x, 'max_x': max_x, 'min_z': min_z, 'max_z': max_z}

        return vol_surface_dict, extremes_dict

    def _get_tenor_index(self, tenor):
        return self._tenors.index(tenor)

    def _get_tenor_expiry(self, tenor):
        return

    def _findate(self, date):
        return Date(date.day, date.month, date.year)


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro - https://www.cuemacro.com / @cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import numpy as np
import pandas as pd

from numba import guvectorize

from findatapy.timeseries import Calendar
from finmarketpy.util.marketconstants import MarketConstants
from finmarketpy.curve.abstractpricer import AbstractPricer

market_constants = MarketConstants()

@guvectorize(['void(f8[:], f8[:], f8[:,:], f8[:,:], intp, f8[:])'],
             '(n),(n),(n,m),(n,m),()->(n)', cache=True, target="cpu", nopython=True)
def _forwards_interpolate_numba(spot_arr, spot_delivery_days_arr, quoted_delivery_days_arr,
                          forwards_points_arr, no_of_tenors, out):

    for i, delivery_day in enumerate(spot_delivery_days_arr):
        for j in range(no_of_tenors):

            # curr_forward_pts = forwards_points_arr[i, j]
            # curr_delivery_days = quoted_delivery_days_arr[i, j]
            # next_forward_pts = forwards_points_arr[i, j+1]
            # next_delivery_days = quoted_delivery_days_arr[i, j+1]

            # In other words, SP (Spot)
            if delivery_day == 0:
                out[i] = spot_arr[i]

                break

            # Say if we're in between SP and 1W
            elif delivery_day <= quoted_delivery_days_arr[i, j]:
                out[i] = spot_arr[i] + (forwards_points_arr[i, j] / quoted_delivery_days_arr[i, j]) * delivery_day

                break
            # Eg. if we're in between 1W and 2W
            elif delivery_day >= quoted_delivery_days_arr[i, j] and delivery_day <= quoted_delivery_days_arr[i, j + 1]:

                # Alternative interpolation
                # interpolated_forwards_arr[i] = spot_arr[i] + \
                #                                delivery_day * (forwards_points_arr[i, j+1] - forwards_points_arr[i, j]) \
                #                                / (quoted_delivery_days_arr[i, j+1] - quoted_delivery_days_arr[i, j])

                forward_points_per_day = (forwards_points_arr[i, j + 1] - forwards_points_arr[i, j]) \
                                         / (quoted_delivery_days_arr[i, j + 1] - quoted_delivery_days_arr[i, j])

                out[i] = spot_arr[i] + forward_points_per_day * delivery_day \
                                     + forwards_points_arr[i, j] - (forward_points_per_day * quoted_delivery_days_arr[i, j])

                break


@guvectorize(['void(f8[:], f8[:,:], f8[:,:], f8[:,:], f8, f8, intp, f8[:,:])'],
             '(n),(n,m),(n,m),(n,m),(),(),()->(n,m)', cache=True, target="cpu", nopython=True)
def _infer_base_currency_depo_numba(spot_arr, outright_forwards_arr, depo_arr, quoted_delivery_days_arr,
                                    base_conv, terms_conv, no_of_tenors, out):
    for i in range(len(out)):
        for j in range(no_of_tenors):
            out[i, j] = (((1 + depo_arr[i, j] * (quoted_delivery_days_arr[i, j] / terms_conv))
                                       / (outright_forwards_arr[i, j] / spot_arr[i])) - 1) \
                                     / (quoted_delivery_days_arr[i, j] / base_conv)

@guvectorize(['void(f8[:], f8[:,:], f8[:,:], f8[:,:], f8, f8, intp, f8[:,:])'],
             '(n),(n,m),(n,m),(n,m),(),(),()->(n,m)', cache=True, target="cpu", nopython=True)
def _infer_terms_currency_depo_numba(spot_arr, outright_forwards_arr, depo_arr, quoted_delivery_days_arr,
                                    base_conv, terms_conv, no_of_tenors, out):

    for i in range(len(out)):
         for j in range(no_of_tenors):
             out[i,j] = ((outright_forwards_arr[i,j] / spot_arr[i]) *
                                      (1 + depo_arr[i,j] * (quoted_delivery_days_arr[i,j] / base_conv)) - 1) \
                                     / (quoted_delivery_days_arr[i,j] / terms_conv)

def _forwards_interpolate(spot_arr, spot_delivery_days_arr, quoted_delivery_days_arr,
                          forwards_points_arr, no_of_tenors):

    out = np.zeros(len(quoted_delivery_days_arr)) * np.nan

    for i, delivery_day in enumerate(spot_delivery_days_arr):
        for j in range(no_of_tenors):

            # curr_forward_pts = forwards_points_arr[i, j]
            # curr_delivery_days = quoted_delivery_days_arr[i, j]
            # next_forward_pts = forwards_points_arr[i, j+1]
            # next_delivery_days = quoted_delivery_days_arr[i, j+1]

            # In other words, SP (Spot)
            if delivery_day == 0:
                out[i] = spot_arr[i]

                break

            # Say if we're in between SP and 1W
            elif delivery_day <= quoted_delivery_days_arr[i, j]:
                out[i] = spot_arr[i] + (forwards_points_arr[i, j] / quoted_delivery_days_arr[i, j]) * delivery_day

                break
            # Eg. if we're in between 1W and 2W
            elif delivery_day >= quoted_delivery_days_arr[i, j] and delivery_day <= quoted_delivery_days_arr[i, j + 1]:

                # Alternative interpolation
                # interpolated_forwards_arr[i] = spot_arr[i] + \
                #                                delivery_day * (forwards_points_arr[i, j+1] - forwards_points_arr[i, j]) \
                #                                / (quoted_delivery_days_arr[i, j+1] - quoted_delivery_days_arr[i, j])

                forward_points_per_day = (forwards_points_arr[i, j + 1] - forwards_points_arr[i, j]) \
                                         / (quoted_delivery_days_arr[i, j + 1] - quoted_delivery_days_arr[i, j])

                out[i] = spot_arr[i] + forward_points_per_day * delivery_day \
                                     + forwards_points_arr[i, j] - (forward_points_per_day * quoted_delivery_days_arr[i, j])

                break

    return out

class FXForwardsPricer(AbstractPricer):
    """Prices forwards for odd dates which are not quoted using linear interpolation,
    eg. if we have forward points for 1W and 1M, and spot date but we want to price a 3W forward, or any arbitrary horizon
    date that lies in that interval

    Also calculates the implied deposit rate from FX spot, FX forward points and deposit rate.
    """

    def __init__(self, market_df=None, quoted_delivery_df=None):
        self._calendar = Calendar()
        self._market_df = market_df
        self._quoted_delivery_df = quoted_delivery_df

    def price_instrument(self, cross, horizon_date, delivery_date, option_expiry_date=None, market_df=None, quoted_delivery_df=None,
                         fx_forwards_tenor_for_interpolation=market_constants.fx_forwards_tenor_for_interpolation,
                         return_as_df=True):
        """Creates an interpolated outright FX forward (and the associated points), for horizon dates/delivery dates
        given by the user from FX spot rates and FX forward points. This can be useful when we have an odd/broken date
        which isn't quoted.

        Uses linear interpolation between quoted dates to calculate the appropriate interpolated forward. Eg. if we
        ask for a delivery date in between 1W and 1M, we will interpolate between those.

        Parameters
        ----------
        cross : str
            Currency pair

        horizon_date : DateTimeIndex
            Horizon dates for forward contracts

        delivery_date : DateTimeIndex
            Delivery dates for forward contracts

        market_df : DataFrame
            Contains FX spot and FX forward points data

        quoted_delivery_df : DataFrame (DateTimeIndex)
            Delivery dates for every quoted forward point

        fx_forwards_tenor_for_interpolation : str(list)
            Which forwards to use for interpolation

        Returns
        -------
        DataFrame
        """

        if market_df is None: market_df = self._market_df
        if quoted_delivery_df is None: quoted_delivery_df = self._quoted_delivery_df

        if quoted_delivery_df is None:
            quoted_delivery_df = self.generate_quoted_delivery(cross, market_df, quoted_delivery_df,
                                                               fx_forwards_tenor_for_interpolation, cross)

        # Make horizon date and delivery date pandas DatetimeIndex
        if isinstance(horizon_date, pd.Timestamp):
            horizon_date = pd.DatetimeIndex([horizon_date])
            delivery_date = pd.DatetimeIndex([delivery_date])
        else:
            horizon_date = pd.DatetimeIndex(horizon_date)
            delivery_date = pd.DatetimeIndex(delivery_date)

        # Truncate the market data so covers only the dates where we want to price the forwards
        market_df = market_df[market_df.index.isin(horizon_date)]
        quoted_delivery_df = quoted_delivery_df[quoted_delivery_df.index.isin(horizon_date)]

        cal = cross

        # Get the spot date (different currency pairs have different conventions for this!)
        spot_date = self._calendar.get_spot_date_from_horizon_date(horizon_date, cal)

        # How many days between spot date (typically T+1 or T+2) and delivery date
        spot_delivery_days_arr = (delivery_date - spot_date).days

        spot_arr = market_df[cross + '.close'].values

        quoted_delivery_df, quoted_delivery_days_arr, forwards_points_arr, divisor = \
            self._setup_forwards_calculation(cross, spot_date, market_df, quoted_delivery_df, fx_forwards_tenor_for_interpolation)

        interpolated_outright_forwards_arr = _forwards_interpolate_numba(spot_arr, spot_delivery_days_arr, quoted_delivery_days_arr,
                          forwards_points_arr, len(fx_forwards_tenor_for_interpolation))

        if return_as_df:
            interpolated_df = pd.DataFrame(index=market_df.index,
                                columns=[cross + '-interpolated-outright-forward.close', cross + "-interpolated-forward-points.close"])

            interpolated_df[cross + '-interpolated-outright-forward.close'] = interpolated_outright_forwards_arr
            interpolated_df[cross + "-interpolated-forward-points.close"] = (interpolated_outright_forwards_arr - spot_arr) * divisor

            return interpolated_df

        return interpolated_outright_forwards_arr

    def get_day_count_conv(self, currency):
        if currency in market_constants.currencies_with_365_basis:
            return 365.0

        return 360.0

    def get_forwards_divisor(self, currency):
        # Typically most divisors of forward points are 10000.0
        divisor = 10000.0

        if currency in market_constants.fx_forwards_points_divisor_1:
            divisor = 1.0

        elif currency in market_constants.fx_forwards_points_divisor_100:
            divisor = 100.0

        elif currency in market_constants.fx_forwards_points_divisor_1000:
            divisor = 1000.0

        return divisor

    def _setup_forwards_calculation(self, cross, spot_date, market_df, quoted_delivery_df, fx_forwards_tenor):

        cal = cross

        # Get the quoted delivery dates for every quoted tenor in our forwards market data
        # Eg. what's the delivery date for EURUSD SN, 1W etc.
        quoted_delivery_df = self.generate_quoted_delivery(cross, market_df, quoted_delivery_df,
                                                           fx_forwards_tenor, cal)

        # Typically most divisors of forward points are 10000.0 (but JPY is an exception)
        divisor = self.get_forwards_divisor(cross[3:6])

        forwards_points_arr = market_df[
                                  [cross + x + ".close" for x in fx_forwards_tenor]].values / divisor

        quoted_delivery_days_arr = np.zeros((len(quoted_delivery_df.index), len(fx_forwards_tenor)))

        # Get difference between each quoted point and the spot date (NOT horizon date)
        # eg. for 1W this is generally 7 days, for 2W it's 14 days etc.
        for i, tenor in enumerate(fx_forwards_tenor):
            quoted_delivery_days_arr[:, i] = (
                        pd.DatetimeIndex(quoted_delivery_df[cross + tenor + ".delivery"]) - spot_date).days

        return quoted_delivery_df, quoted_delivery_days_arr, forwards_points_arr, divisor

    def generate_quoted_delivery(self, cross, market_df, quoted_delivery_df, fx_forwards_tenor, cal):

        if not(isinstance(fx_forwards_tenor, list)):
            fx_forwards_tenor = [fx_forwards_tenor]

        # Get the quoted delivery dates for every quoted tenor in our forwards market data
        # Eg. what's the delivery date for EURUSD SN, 1W etc.
        if quoted_delivery_df is None:
            quoted_delivery_df = pd.DataFrame(index=market_df.index,
                                              columns=[cross + tenor + ".delivery" for tenor in
                                                       fx_forwards_tenor])


            for tenor in fx_forwards_tenor:
                quoted_delivery_df[cross + tenor + ".delivery"] = \
                    self._calendar.get_delivery_date_from_horizon_date(quoted_delivery_df.index, tenor, cal=cal)

        return quoted_delivery_df

    def calculate_implied_depo(self, cross, implied_currency, market_df=None, quoted_delivery_df=None,
                               fx_forwards_tenor=market_constants.fx_forwards_trading_tenor,
                               depo_tenor=None):
        """Calculates implied deposit rates for a particular currency from spot, forward points and deposit rate
        for the other currency. Uses the theory of covered interest rate parity.

        See BIS publication from 2016, Covered interest parity lost: understanding the cross-currency basis downloadable from
        https://www.bis.org/publ/qtrpdf/r_qt1609e.pdf for an explanation

        Eg. using EURUSD spot, EURUSD 1W forward points and USDON deposit rate, we can calculate the implied deposit
        for EUR 1W

        Parameters
        ----------
        cross : str
            Currency pair

        implied_currency : str
            Currency for which we want to imply deposit

        market_df : DataFrame
            With FX spot rate, FX forward points and deposit rates

        fx_forwards_tenor : str (list)
            Tenors of forwards where we want to imply deposit

        depo_tenor : str
            Deposit rate to use (default - ON)

        Returns
        -------
        DataFrame
        """

        if not(isinstance(fx_forwards_tenor, list)):
            fx_forwards_tenor = [fx_forwards_tenor]

        if market_df is None: market_df = self._market_df
        if quoted_delivery_df is None: quoted_delivery_df = self._quoted_delivery_df
        if depo_tenor is None: depo_tenor = fx_forwards_tenor

        if not(isinstance(depo_tenor, list)):
            depo_tenor = [depo_tenor] * len(fx_forwards_tenor)

        cal = cross

        # Get the spot date (different currency pairs have different conventions for this!)
        spot_date = self._calendar.get_spot_date_from_horizon_date(market_df.index, cal)

        quoted_delivery_df, quoted_delivery_days_arr, forwards_points_arr, divisor = \
            self._setup_forwards_calculation(cross, spot_date, market_df, quoted_delivery_df,
                                             fx_forwards_tenor)

        spot_arr = market_df[cross + '.close'].values

        outright_forwards_arr = np.vstack([spot_arr]*len(fx_forwards_tenor)).T + forwards_points_arr

        base_conv = self.get_day_count_conv(cross[0:3])
        terms_conv = self.get_day_count_conv(cross[3:6])

        # Infer base currency
        if implied_currency == cross[0:3]:
            original_currency = cross[3:6]

            depo_arr = market_df[[original_currency + d + '.close' for d in depo_tenor]].values / 100.0

            implied_depo_arr = _infer_base_currency_depo_numba(spot_arr, outright_forwards_arr, depo_arr, quoted_delivery_days_arr,
                                    base_conv, terms_conv, len(fx_forwards_tenor))

            # for i in range(len(implied_depo_arr)):
            #     for j, tenor in enumerate(fx_forwards_tenor):
            #         implied_depo_arr[i, j] = (((1 + depo_arr[i,j] * (quoted_delivery_days_arr[i, j] / terms_conv))
            #                                    / (outright_forwards_arr[i,j] / spot_arr[i])) - 1) \
            #                                  / (quoted_delivery_days_arr[i, j] / base_conv)

        # Infer terms currency
        if implied_currency == cross[3:6]:
            original_currency = cross[0:3]

            depo_arr = market_df[[original_currency + d + '.close' for d in depo_tenor]].values / 100.0

            implied_depo_arr = _infer_terms_currency_depo_numba(spot_arr, outright_forwards_arr, depo_arr, quoted_delivery_days_arr,
                                    base_conv, terms_conv, len(fx_forwards_tenor))

            # for i in range(len(implied_depo_arr)):
            #     for j, tenor in enumerate(fx_forwards_tenor):
            #         implied_depo_arr[i,j] = ((outright_forwards_arr[i,j] / spot_arr[i]) *
            #                                  (1 + depo_arr[i,j] * (quoted_delivery_days_arr[i,j] / base_conv)) - 1) \
            #                                 / (quoted_delivery_days_arr[i,j] / terms_conv)

        return pd.DataFrame(index=market_df.index,
                            columns=[implied_currency + x + "-implied-depo.close" for x in fx_forwards_tenor],
                            data=implied_depo_arr * 100.0)


__author__ = "saeedamen"

#
# Copyright 2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#


# loading data
import datetime

from chartpy import Chart, Style
from findatapy.market import Market, MarketDataGenerator, MarketDataRequest
from finmarketpy.economics import TechIndicator, TechParams

from findatapy.util.loggermanager import LoggerManager

logger = LoggerManager().getLogger(__name__)

chart = Chart(engine="matplotlib")

market = Market(market_data_generator=MarketDataGenerator())
tech_ind = TechIndicator()

# Choose run_example = 0 for everything
# run_example = 1 - download S&P500 from Quandl, calculate ATR and plot

run_example = 0

###### Fetch data from Quandl for BoE rate (using Bloomberg data)
if run_example == 1 or run_example == 0:

    # Downloaded S&P500
    md_request = MarketDataRequest(
                start_date = "01 Jan 2000",                         # start date
                data_source = "quandl",                             # use Quandl as data source
                tickers = ["S&P500"],
                fields = ["close", "open", "high", "low"],          # which fields to download
                vendor_tickers = ["YAHOO/INDEX_GSPC"],              # ticker (Bloomberg)
                vendor_fields = ["close", "open", "high", "low"],   # which Bloomberg fields to download
                cache_algo = "internet_load_return")                # how to return data

    df = market.fetch_market(md_request)

    print(df)

    tech_params = TechParams()
    tech_params.atr_period = 14
    tech_ind.create_tech_ind(df, "ATR", tech_params)

    style = Style()

    style.title = "S&P500 ATR"
    style.scale_factor = 2
    style.file_output = "sp500.png"
    style.source = "Quandl/Yahoo"

    df = tech_ind.get_techind()

    chart.plot(df, style=style)
